/*
 * blk_tree.cpp
 * Copyright (C) 2021 Aitor Cuadrado Zubizarreta <aitor_czr@gnuinos.org>
 * 
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * See the LICENSE file. * 
 */


#include "blk_tree.h"
#include "hopman_ui.h"
#include "fpipe.h"
#include "eject_dialog.h"
#include "svr.h"

namespace Hopman {
    
BlkTree::BlkTree(const Hopman::Annex& annex)
: m_WorkerThread (nullptr),
  display_menu (false),
  m_shall_quit (true)
{
    m_tooltip_window = new Gtk::Window(Gtk::WINDOW_POPUP);
    m_tooltip_window->set_default_size(0,0);
    set_tooltip_window(*m_tooltip_window);
    m_tooltip_window->add(m_Label);
    m_tooltip_window->show_all_children();
    
    get_selection()->set_mode( Gtk::SELECTION_SINGLE );
    set_hover_selection (true);
    set_sensitive (true);
    property_has_tooltip();

    add_events(Gdk::BUTTON_PRESS_MASK); 
    
#if GTK_MAJOR_VERSION == 3

    set_activate_on_single_click( true );
        
#endif
  
    set_headers_visible (false);
    set_enable_tree_lines (true);
    //set_grid_lines(Gtk::TREE_VIEW_GRID_LINES_BOTH);
  
    m_refTreeModel = Gtk::TreeStore::create( m_Columns );    
    set_model( m_refTreeModel );
  
    append_column( "DEVICE", m_Columns.device );
    append_column( "FSTYPE", m_Columns.fstype );
    append_column( "SIZE", m_Columns.size );
    append_column( "ICON", m_Columns.icon );
    append_column( "VENDOR", m_Columns.vendor );
    
    m_Menu = new cMenu(this);
    selected.clear();

    show_all_children();
    
    signal_query_tooltip().connect
    (
        sigc::mem_fun(*this, &BlkTree::on_query_tooltip),
        false
    );
    
    signal_button_press_event().connect
    (
        sigc::mem_fun(*this, &BlkTree::on_button_press_event),
        false
    );
    
    m_Dispatcher.connect
    (
        sigc::bind<const Hopman::Annex&>
        (
            sigc::mem_fun(*this, &BlkTree::get_info_from_worker_thread),
            annex
        )
    );
 
    // Start a new worker thread.
    if (!m_WorkerThread)
    {
        m_WorkerThread = new thread (
        [ this, annex ]
            {
                m_Worker.do_work(this, annex);
            }
        );
    }
    
    set_can_default();
    show_all_children();
    grab_focus();
    
    this->removed_icon = annex.removed_icon;
    this->socket = annex.socket; 
    this->path_to_the_helper = annex.path_to_the_helper;
}

// Destructor
BlkTree::~BlkTree()
{
    delete m_tooltip_window;
    delete m_Menu;
    
    // Order the worker thread to stop and wait for it to stop.
    if (m_WorkerThread->joinable())
    {
        m_WorkerThread->detach();
        m_Worker.stop_work();
        if(m_WorkerThread) {
            delete m_WorkerThread;
            m_WorkerThread = nullptr;
        }
    }
                
    // Purge the contents
    pObj.clear();

    // Deallocate memory
    pObj.shrink_to_fit();
}

bool BlkTree::get_shall_quit()
{
    return this->m_shall_quit;
}

string BlkTree::get_selected()
{
    return selected;
}

void BlkTree::notify()
{
    m_Dispatcher.emit();
}

void BlkTree::get_info_from_worker_thread(const Hopman::Annex& annex)
{
    Hopman::Hopman_Ui *parent = dynamic_cast<Hopman::Hopman_Ui*>(this->get_toplevel());
    if (parent) {
        parent->resize(5,5);
        parent->m_Spinner.stop();
        parent->m_Spinner.hide();
        parent->m_Label.hide();
    }

    pObj.clear();
    pObj.begin();
    m_Worker.get_data( pObj );
   
    // No devices?    
    if(!pObj.size())
     {  
        m_refTreeModel->clear();
        parent->resize(5,5);  
        Gtk::MessageDialog dialog(*parent, "No devices found.");
        dialog.run();
        gtk_main_quit();
    }
   
    // Fill the tree    
    m_refTreeModel->clear();
    Glib::RefPtr<Gdk::Pixbuf> warning_icon = Gdk::Pixbuf::create_from_file( annex.warning_icon );    
    Glib::RefPtr<Gdk::Pixbuf> removed_icon = Gdk::Pixbuf::create_from_file( annex.removed_icon );    
    Glib::RefPtr<Gdk::Pixbuf> eject_icon = Gdk::Pixbuf::create_from_file( annex.eject_icon );
    for(auto &u : pObj) {
        row = *(m_refTreeModel->append());
        row[m_Columns.device] = " " + u->get_devname();
        row[m_Columns.fstype] = " " + u->get_fstype();
        string str = u->get_size() + " G";
        int n = str.length();
        for(unsigned i=n; i<8; i++) str = " " + str;
        row[m_Columns.size] = str + "  ";
        row[m_Columns.vendor] = "  [ " + u->get_vendor() + " - " + u->get_model() +  " ] ";
        bool mounted=false;
        int num_children = 0;
        for(auto &w : u->children) {
            childrow = *(m_refTreeModel->append(row.children()));
            childrow[m_Columns.device] = w->get_devname();
            childrow[m_Columns.fstype] = "  " + w->get_fstype();
            string str = w->get_size() + " G";
            int n = str.length();
            for(unsigned i=n; i<8; i++) str = " " + str;
            childrow[m_Columns.size] = str + "  ";
            if(w->get_mtpt().length()) {
                childrow[m_Columns.vendor] = "  >>  " + w->get_mtpt();
                childrow[m_Columns.icon] = eject_icon;
                mounted=true;
            }
            num_children++;
        }
        
        if(mounted) row[m_Columns.icon] = warning_icon;
        else {
            if(!u->get_removable()) row[m_Columns.icon] = removed_icon;
            else row[m_Columns.icon] = eject_icon;
        }
    }
   
    // Set parent's gravity
    int x, y;
    parent->get_position(x, y);
    if(x > annex.x_screen/2) {
        if(y > annex.y_screen/2) parent->set_gravity(Gdk::GRAVITY_SOUTH_EAST);
        else parent->set_gravity(Gdk::GRAVITY_NORTH_EAST); 
    } else {
        if(y > annex.y_screen/2) parent->set_gravity(Gdk::GRAVITY_SOUTH_WEST);
        else parent->set_gravity(Gdk::GRAVITY_NORTH_WEST);
    } 
/*    
    while (Gtk::Main::events_pending ())
        Gtk::Main::iteration (); 
        * */ 
}

// Signal Handler (single click)
bool BlkTree::on_button_press_event(GdkEventButton* event)
{
    //Call base class, to allow normal handling (requires: set_hover_selection( true ) );
    Gtk::TreeView::on_button_press_event(event);
      
    if(event->type == GDK_BUTTON_PRESS)
    {
        if(event->button == 1)
            return on_left_button_press_event(event);
        else if(event->button == 2)
            return on_middle_button_press_event(event);
        else if(event->button == 3)
            return on_right_button_press_event(event);
    }
    else if(event->type==GDK_2BUTTON_PRESS)
        return on_double_click_event(event);
            
    return false;
}
    
bool BlkTree::on_left_button_press_event(GdkEventButton* event)
{
    if(!this->do_action) return false;
    this->do_action = false;
        
    Glib::RefPtr<Gtk::TreeView::Selection> refTreeSelection = get_selection();
    if(refTreeSelection)
    {
        Gtk::TreeModel::iterator iter = refTreeSelection->get_selected();
        if(iter) {
			if( (*iter).parent() ) {
				string str = (*iter)[ m_Columns.device ];
				string cmd = "pumount " + str + " 2>&1";
    
				FILE *fp;
				char buf[1024];
				string e( "" );
            
				fp = popen( cmd.c_str(), "r" );
				if(!fp) {
					cout << "Popen error\n";
					exit( EXIT_FAILURE );
				}
    
				// There is a bug in pmount related to the replicated stdout error message
				string first_line( "" );
				while(fgets(buf, sizeof(buf), fp)) {
					if( !first_line.length() ) first_line = string(buf);
					e+=string(buf);
				}
    
				// Is it replicated? If so, split it half
				size_t pos=e.find(first_line, 2);
				if(pos != -1) {
					string str1=e.substr(0, pos);
					string str2=e.substr(pos);
					if(!str2.compare(str1)) e=str1;
				}
				
				if(WEXITSTATUS(pclose(fp)) == 5) {
					Gtk::Window *parent = dynamic_cast<Gtk::Window*>(this->get_toplevel());
					Gtk::MessageDialog dialog(*parent, "This is an INFO MessageDialog");
					dialog.set_message( "[ERROR]: umount failed", false );
					dialog.set_secondary_text( e.c_str(), false );
					dialog.run();
				}
			}
			else {
				bool is_device_unmounted = true;
				Gtk::TreeNodeChildren childList = (*iter).children();
				Gtk::TreeNodeChildren::iterator it = childList.begin();
				int n=0;
				while(n < childList.size()) {
					n++;
					string mtpt = (*it)[ m_Columns.vendor ];
					if( mtpt.length() )
						is_device_unmounted = false;
					it++;
				}
				
				if(is_device_unmounted)
					on_double_click_event(event);
			}
		}
    }
    
    return false;
}

bool BlkTree::on_double_click_event(GdkEventButton* event)
{    
    Glib::RefPtr<Gtk::TreeView::Selection> refTreeSelection = get_selection();
    if(refTreeSelection)
    {
        Gtk::TreeModel::iterator iter = refTreeSelection->get_selected();
        string devname = (*iter)[ m_Columns.device ];
		// Remove spaces in blanck;
		devname.erase(std::remove(devname.begin(), devname.end(), ' '), devname.end());
        bool removable = true;
        for(auto &u : pObj)
			if( !(u->get_devname()).compare(devname) && !u->get_removable() )
				removable = false;
		
        if(removable) {
			this->m_shall_quit = false;
    
			Hopman::EjectDialog *eject_dialog = new Hopman::EjectDialog( this->selected, this->removed_icon, this->socket, this->path_to_the_helper );
			eject_dialog->run();
			delete eject_dialog;
    
			this->m_shall_quit = true;
		}
		else {
           Gtk::Window *parent = dynamic_cast<Gtk::Window*>(this->get_toplevel());
           Gtk::MessageDialog dialog(*parent, "This is an INFO MessageDialog");
           dialog.set_message( "[INFO]: This device isn't removable", false );
           dialog.run();
        }
	}
    
    return true;
}

bool BlkTree::on_middle_button_press_event(GdkEventButton* event)
{
    /* *
     *    
     * */
     
    return true;
}

bool BlkTree::on_right_button_press_event(GdkEventButton* event)
{
    if(display_menu) {
		
        m_Menu->display_menu( event, selected, BlkTree::get_mountpoint( selected ) );
        return true;
    } else {
        
        return false;
    }
}

bool BlkTree::is_device_removable(shared_ptr<Hopman::BlkNode> node)
{
    try {
        // Read the value in $SYSFS_PATH/removable
        string filename = node->get_pathname() + "/removable";
        unique_ptr<FILE, decltype(&fclose)> fp(::fopen(filename.c_str(), "r"), &fclose);
        char ch = (char)fgetc(fp.get());
        if (ch == '0')
            return false;
        else if (ch == '1')
            return true;
    }    
    
    catch(exception const& e) {
            
        throw runtime_error("Sorry, there was an error opening the pipe\n");        
    };
}

string BlkTree::get_mountpoint(string partition)
{

    
    /*
    string dev=partition;
    dev.pop_back();
    for(auto &u : pObj) {
        if(!dev.compare(u->get_devname()))
            for(auto &v : u->children) {
                if(!v->get_devname().compare(partition) && v->get_mtpt().length()) {
                    res = false;
                }
            }
    }
    * */
    
    
                DIR *dirp;
            struct dirent *E;
        
            vector<string> procmounts;
            procmounts.begin();
            try {
        
                char ch;
                string str;        
                unique_ptr<FILE, decltype(&fclose)> fp(::fopen("/proc/mounts", "r"), &fclose);
                while ((ch = (char)fgetc(fp.get()))!=EOF) {
                    if(ch != '\n') {
                        stringstream s_stream;
                        s_stream << ch;
                        str.append(s_stream.str());
                    }
                    else {
                        procmounts.push_back(str);
                        str.clear();
                    }
                }
            }
    
            catch(exception const& e) {
        
                throw runtime_error("Sorry, there was an error opening the pipe\n");        
            };
    
    
    
                                        
                    string mountpoint( "" );
                    for(auto &u : procmounts) {
                        vector<string> tokens=BlkTree::split_line(u, " ");
                        if(!tokens[0].compare( partition )) {
                            mountpoint=tokens[1];
                        }
                    }
    
    return mountpoint;
}
    
void BlkTree::on_mount()
{
    FILE *fp;
    char buf[1024];
    string e( "" );
    string cmd = "pmount " + selected + " 2>&1";
    
    fp = popen( cmd.c_str(), "r" );
    if(!fp) {
        cout << "Popen error\n";
        exit( EXIT_FAILURE );
    }
    
    // There is a bug in pmount related to the replicated stdout error message
    string first_line( "" );
    while(fgets(buf, sizeof(buf), fp)) {
        if( !first_line.length() ) first_line = string(buf);
        e+=string(buf);
    }
    
    // Is it replicated? If so, split it half
    size_t pos=e.find(first_line, 2);
    if(pos != -1) {
        string str1=e.substr(0, pos);
        string str2=e.substr(pos);
        if(!str2.compare(str1)) e=str1;
    }

    if( WEXITSTATUS(pclose(fp)) != 255) {
        Gtk::Window *parent = dynamic_cast<Gtk::Window*>(this->get_toplevel());
        Gtk::MessageDialog dialog(*parent, "[ERROR]: pmount failed");
        dialog.set_secondary_text( e.c_str(), false );
        dialog.run();                
        gtk_main_quit();
    }
}

// Signal Handler
void BlkTree::on_row_activated(const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn *column)
{
    if(column->get_title().compare("ICON")) return;
    else this->do_action = true;
}

void BlkTree::on_cursor_changed()
{
    display_menu = false;
    selected.clear();
    Glib::RefPtr<Gtk::TreeView::Selection> refTreeSelection = get_selection();
    if(refTreeSelection)
    {
        Gtk::TreeModel::iterator iter = refTreeSelection->get_selected();
        if(iter) {
            string strLabel( "" );
            string temp( "" );
            
            string str=(*iter)[ m_Columns.device ];
            str.erase(std::remove(str.begin(), str.end(), ' '), str.end());   /*  Remove spaces in blanck  */
            this->selected  = str;
            
            auto node = BlkTree::getNode(this->selected);
                
            if( (*iter).parent() ) {     /* child */
                                
                // In  the children, we'll specify parent's label too                            
                // Parent's weak_ptr has to be copied into a shared_ptr before usage                        
                if (auto p = node->parent.lock()) {
                    
                    display_menu = true;
                    temp=p->get_label();
                    if(temp.size()) temp+=" | ";
                    
                } else {
                     
                    throw runtime_error( "weak pointer is expired\n" );
                }

                temp+=node->get_label();
                if(temp.size()) strLabel+=" LABEL=" + temp + " \n";                
                strLabel+=" UUID=" + node->get_uuid() + " \n";
                strLabel+=" PARTUUID=" + node->get_partuuid() + " ";
                
            } else {    /* parent */
                                
                temp=node->get_label();
                if(temp.size()) strLabel+=" LABEL=" + temp + " \n";
                temp=node->get_uuid();
                if(temp.size()) strLabel+=" UUID=" + node->get_uuid() + " ";                
            }
            
            m_Label.set_text( strLabel );
        }
    }
}

bool BlkTree::on_test_collapse_row (const Gtk::TreeModel::iterator& iter, const Gtk::TreeModel::Path& path)
{
    Gtk::Window *parent = dynamic_cast<Gtk::Window*>(this->get_toplevel());
    if (parent) 
        parent->resize(5,5);
    return Gtk::TreeView::on_test_collapse_row (iter, path);
}

bool BlkTree::on_query_tooltip(int x, int y, bool keyboard_tooltip, const Glib::RefPtr<Gtk::Tooltip>& tooltip)
{
    if( m_Label.get_text().length() > 1) m_Label.show();
    m_tooltip_window->hide();
    m_tooltip_window->resize(1, 1);
    
    return true;
}

shared_ptr<Hopman::BlkNode> BlkTree::getNode( string dev )
{
    string duplicate = dev;
    duplicate.pop_back();
    
    for(auto &p : pObj) {        
        string d_name=p->get_devname();
        if( !dev.compare(d_name) )
            return p;
        else if( !duplicate.compare(d_name) )
            for(auto &w : p->children)
                if( !w->get_devname().compare(dev) )
                    return w;
    }
    
    return nullptr;
}

vector<string> BlkTree::split_line(const string& str, const string& delim)
{
    vector<string> tokens;
    size_t prev = 0, pos = 0;
    do {
        pos = str.find(delim, prev);
        if(pos == string::npos)
            pos = str.length();
        string token = str.substr(prev, pos-prev);
        if(!token.empty())
            tokens.push_back(token);
        prev = pos + delim.length();
    } while (pos < str.length() && prev < str.length());
    
    return tokens;
}

} //namespace Hopman
