/*
 * blk_tree.h
 * Copyright (C) 2021 Aitor Cuadrado Zubizarreta <aitor_czr@gnuinos.org>
 * 
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * See the LICENSE file. * 
 */


#ifndef __BLK_TREE_H__
#define __BLK_TREE_H__

#include <gtkmm.h>
#include <thread>
#include <vector>

#include "annex.h"
#include "blk_model.h"
#include "blk_node.h"

namespace Hopman {
    
using namespace std;
    
class BlkTree : public Gtk::TreeView
{
public:
    BlkTree(const Hopman::Annex&);
    virtual ~BlkTree ();
    
    bool get_shall_quit();

    class cMenu : public Gtk::Menu
    {
    public:
        cMenu(BlkTree*);
        virtual ~cMenu();
    
        void display_menu(GdkEventButton*, string, string);
  
    private:
		string name, mtpt;
		
        // Child widgets
        Gtk::MenuItem *item;
    
        virtual void on_open_new_tab(BlkTree*);
        virtual void on_open_new_window(BlkTree*);
    };
  
    string get_selected();
    
    class Worker
    {
        Worker();
        virtual ~Worker();
        
    private:
		int inot_fd, watch_dirs;
		
        struct BlkData { // data to manage 
            BlkData( ) { }  
            virtual ~BlkData() { }
            string path, fstype, label, uuid, partuuid;
        };
        
        vector< shared_ptr<Hopman::BlkNode> > pObj_Worker;
        void do_work(BlkTree*, const Hopman::Annex&);         
        void stop_work();
        void get_data(vector< shared_ptr<Hopman::BlkNode> >&) const;
        static bool is_partition(string);
        static string get_basename(string);
        static string rm_digit(string);
        static bool compareNames(shared_ptr<Hopman::BlkNode>, shared_ptr<Hopman::BlkNode>);
        static bool compareNodes(shared_ptr<Hopman::BlkNode>, shared_ptr<Hopman::BlkNode>);
        static vector<string> split_line(const string&, const string&);
        static void showList(struct vdev_usb_device_list*);
    
        // Synchronizes access to member data.
        mutable mutex m_Mutex;
  
        // Data used by both GUI thread and worker thread.
        bool m_first_time, m_shall_stop;
        
        // Declare the outer class as friend of the inner class:
        friend class BlkTree;
    };

protected:
    bool on_button_press_event(GdkEventButton*) override;
    bool on_left_button_press_event(GdkEventButton*);
    bool on_right_button_press_event(GdkEventButton*);
    bool on_middle_button_press_event(GdkEventButton*);
    bool on_double_click_event(GdkEventButton*);    
    void on_mount();
    string get_mountpoint(string);
    static vector<string> split_line(const string&, const string&);
    
    /* This is for the double click event */
    void on_row_activated(const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn* /* column */);
    bool on_test_collapse_row (const Gtk::TreeModel::iterator&, const Gtk::TreeModel::Path&);
    void on_cursor_changed();
    
    /* Tooltip */
    bool on_query_tooltip(int, int, bool, const Glib::RefPtr<Gtk::Tooltip>&);
  
private:
    vector< shared_ptr<Hopman::BlkNode> > pObj;
    Gtk::Window *m_tooltip_window;
    Gtk::Label m_Label;
    bool display_menu, do_action, m_shall_quit;
    string selected, removed_icon, socket, path_to_the_helper;
    
    Worker m_Worker;
    thread *m_WorkerThread;
    
    // Glib:
    Glib::Dispatcher m_Dispatcher;
    
    // Called from the worker thread:
    void fill_tree();
    void notify();
    
    // Dispatcher handler:
    void get_info_from_worker_thread(const Hopman::Annex&);
    
    bool is_device_removable(shared_ptr<Hopman::BlkNode>);
  
    cMenu *m_Menu;
    guint cols_count;
    Hopman::BlkModel m_Columns;
    Glib::RefPtr<Gtk::TreeStore> m_refTreeModel;         
    Gtk::TreeModel::Row row, childrow;
    Gtk::TreeViewColumn* pColumn;
    Gtk::CellRendererProgress* cell;
    
    shared_ptr<Hopman::BlkNode> getNode(string);
};

typedef Hopman::BlkTree::cMenu BlkTreeCMenu;
typedef Hopman::BlkTree::Worker BlkTreeWorker;

} // namespace Hopman

#endif // __BLK_TREE_H__

