/*
 * eject_dialog.cpp
 * Copyright (C) 2021 Aitor Cuadrado Zubizarreta <aitor_czr@gnuinos.org>
 * 
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * See the LICENSE file. * 
 */
 

#include "eject_dialog.h"
#include "buttonbox.h"
#include "blk_tree.h"
#include "fpipe.h"
#include "svr.h"

#include <sys/inotify.h>

#define EVENT_SIZE  ( sizeof (struct inotify_event) )
#define EVENT_BUF_LEN     ( 1024 * ( EVENT_SIZE + 16 ) )

namespace Hopman {
	
EjectDialog::EjectDialog( string devname, string icon_path, string socket_path, string eject_path )
: m_bActivityMode(true),
  m_WorkerThread(nullptr)
{ 
    set_size_request( 400, 120 );
    set_position( Gtk::WIN_POS_CENTER );
    present();
	
    Gtk::VBox *m_VBox = Gtk::manage( new Gtk::VBox() );
	
#if GTK_MAJOR_VERSION == 2

    m_VBox = get_vbox();
	
#elif GTK_MAJOR_VERSION == 3

    m_VBox = get_content_area();
		
#endif

    m_VBox->add(m_Align1);
    Gtk::HBox *hBox1 = new Gtk::HBox;
    m_Align1.add(*hBox1);

    hBox1->set_border_width(10);
	
    hBox1->add(m_Image);
    hBox1->pack_start(m_Label, Gtk::PACK_EXPAND_WIDGET);
    m_Label.set_padding(10, 5);
	
    m_Image.set( icon_path );
    string label = devname + " is going to be removed. Continue?";
    m_Label.set_text( label );

    m_VBox->add(m_Align2);
    Gtk::HBox *hBox2 = new Gtk::HBox;
    m_Align2.add(*hBox2);

    hBox2->set_border_width(0);
 
    myButtonBox *pButtonBox = new myButtonBox(true, 10, Gtk::BUTTONBOX_END);
    hBox2->pack_start(*pButtonBox, true, true, 5);
    
    m_Button_Cancel = Gtk::manage( new Gtk::Button( Gtk::Stock::CANCEL ) );
    m_Button_Cancel->signal_clicked().connect([this] {this->hide();});
    pButtonBox->bbox->add(*m_Button_Cancel);    
    
    m_Button_Eject = Gtk::manage( new Gtk::Button( Gtk::Stock::REMOVE ) ); 
    m_Button_Eject->signal_clicked().connect([this] {this->on_eject();});
    pButtonBox->bbox->add(*m_Button_Eject);
  
    m_VBox->pack_start(m_ProgressBar, Gtk::PACK_SHRINK, 5);
  
    m_ProgressBar.pulse();
 
    /* Add a timer callback to update the value of the progress bar */
    m_connection_timeout = Glib::signal_timeout().connect
    (
        sigc::mem_fun(*this, &EjectDialog::on_timeout),
        50
    );
  
    op_client = sigc::mem_fun( *this, &EjectDialog::on_eject);
    
    m_Button_Eject->set_can_default(); 
    show_all_children();
    m_Button_Eject->grab_focus();
    m_ProgressBar.hide();
    
    strDevice = devname;
    strSocket = socket_path;
    strEject = eject_path;
}

EjectDialog::~EjectDialog()
{
    if (m_WorkerThread) {
        m_WorkerThread->detach();
        delete m_WorkerThread;
        m_WorkerThread = nullptr;
    }
}

void EjectDialog::on_eject()
{	
    m_Button_Eject->set_sensitive( false );
	
    if (!m_WorkerThread) {
        m_WorkerThread = new thread (
            [ this ] {
				int inot_fd, watch_dirs;int length, count = 0;
				char buffer[EVENT_BUF_LEN];				
	
				/*creating the INOTIFY instance*/
				inot_fd = inotify_init();

				/*checking for error*/
				if ( inot_fd < 0 ) {
					perror( "inotify_init" );
				}
    
				/*adding the “/tmp” directory into watch list. Here, the suggestion is to validate the existence of the directory before adding into monitoring list.*/
				watch_dirs = inotify_add_watch( inot_fd, "/dev/disk", IN_CREATE | IN_DELETE );
				watch_dirs = inotify_add_watch( inot_fd, "/dev/disk/by-id", IN_CREATE | IN_DELETE );
				watch_dirs = inotify_add_watch( inot_fd, "/dev/disk/by-uuid", IN_CREATE | IN_DELETE );
				watch_dirs = inotify_add_watch( inot_fd, "/dev/disk/by-label", IN_CREATE | IN_DELETE );
				
                m_ProgressBar.show();
                    {
                        unique_ptr<Svr> svr( new Svr( strSocket ) );
                        svr->dispatch( { "1", strDevice } );
                        string cmd = strEject + " " + strSocket;
                        system( cmd.c_str() );
                        this_thread::sleep_for(chrono::milliseconds(2000));
                    }

				/*read to determine the event change happens on the monitoring list. Actually this read blocks until the change event occurs*/ 	
				length = read( inot_fd, buffer, EVENT_BUF_LEN );  

				/*checking for error*/
				if ( length < 0 ) {
					perror( "read" );
				}  

				/*actually read return the list of change events happens. Here, read the change event one by one and process it accordingly.*/
				while ( count < length ) {     
					struct inotify_event *event = ( struct inotify_event * ) &buffer[ count ];     
					if ( event->len ) {
						if ( event->mask & IN_CREATE ) {
							if ( event->mask & IN_ISDIR ) {
								printf( "New directory %s created.\n", event->name );
							}
							else {
								printf( "New file %s created.\n", event->name );
							}
						}
						else if ( event->mask & IN_DELETE ) {
							if ( event->mask & IN_ISDIR ) {
								printf( "Directory %s deleted.\n", event->name );
							}
							else {
								printf( "File %s deleted.\n", event->name );
							}
						}
						else if ( event->mask & IN_MODIFY ) {
							printf( "Changed.\n", event->name );
						}
					}
			
					count += EVENT_SIZE + event->len;
				}
	
				/*removing directories from the watch list.*/
				inotify_rm_watch( inot_fd, watch_dirs );
	
				/*closing the INOTIFY instance*/
				close( inot_fd );
                
                m_ProgressBar.hide();
                this->set_size_request( 400, 120 );
                this_thread::sleep_for(chrono::milliseconds(1000));
                hide();
            }
        );
    }
}

bool EjectDialog::on_timeout()
{
    if(m_bActivityMode) m_ProgressBar.pulse();
    else {
        double new_val = m_ProgressBar.get_fraction() + 0.01;

        if(new_val > 1.0) new_val = 0.0;

        /* Set the new value: */
        m_ProgressBar.set_fraction(new_val);
    }
    
    return true;
}

} // namespace Hopman
