/*
 * worker.cpp
 * Copyright (C) 2021 Aitor Cuadrado Zubizarreta <aitor_czr@gnuinos.org>
 * 
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * See the LICENSE file. * 
 */


#include <iostream>
#include <thread>
#include <mutex>
#include <memory>
#include <vector>
#include <sstream>      // stringstream
#include <libgen.h>
#include <regex.h>
#include <algorithm>    // sort
#include <filesystem>
#include <sys/inotify.h>

#include "hopman_ui.h"
#include "blk_tree.h"
#include "blk_node.h"
#include "fpipe.h"
#include "svr.h"

#define EVENT_SIZE  ( sizeof (struct inotify_event) )
#define EVENT_BUF_LEN     ( 1024 * ( EVENT_SIZE + 16 ) )

#define handle_error_en(en, msg) \
               do { errno = en; perror(msg); exit(EXIT_FAILURE); } while (0)

extern "C" {

    #include <hopman.h>
    char *devpath(const char *devname, char *namebuf); /* device path in /sys */
}

namespace Hopman {
    
using namespace std;
namespace fs = std::filesystem;

BlkTreeWorker::Worker()
{    
    m_first_time = true;
    m_shall_stop = false;
}

// Destructor (virtual)
BlkTreeWorker::~Worker()
{
    stop_work();
}

void BlkTreeWorker::get_data (
        vector<shared_ptr<Hopman::BlkNode>>& p
    ) const
{
    lock_guard<mutex> lock(m_Mutex);    
        
    for(auto &u : pObj_Worker)
        p.push_back(u);
}

void BlkTreeWorker::stop_work()
{
    lock_guard<mutex> lock(m_Mutex);
    m_shall_stop = true;
}

void BlkTreeWorker::do_work(BlkTree* caller, const Hopman::Annex& annex)
{    
    // The mutex is unlocked here by lock's destructor.
    {
        lock_guard<mutex> lock(m_Mutex);
    }
    
    vector< shared_ptr<Hopman::BlkNode> > copy_obj;
    
    // do until break
    while(1)
    {
        string name, mtpt;
	
		/*creating the INOTIFY instance*/
		inot_fd = inotify_init();

		/*checking for error*/
		if ( inot_fd < 0 ) {
			perror( "inotify_init" );
		}
    
		/*adding the “/tmp” directory into watch list. Here, the suggestion is to validate the existence of the directory before adding into monitoring list.*/
		watch_dirs = inotify_add_watch( inot_fd, "/dev/disk", IN_CREATE | IN_DELETE );
		watch_dirs = inotify_add_watch( inot_fd, "/dev/disk/by-id", IN_CREATE | IN_DELETE );
		watch_dirs = inotify_add_watch( inot_fd, "/dev/disk/by-uuid", IN_CREATE | IN_DELETE );
		watch_dirs = inotify_add_watch( inot_fd, "/dev/disk/by-label", IN_CREATE | IN_DELETE );
		watch_dirs = inotify_add_watch( inot_fd, "/tmp/hopman1000", IN_CREATE | IN_DELETE | IN_MODIFY );
        
        struct vdev_usb_device_list *list = nullptr;
        
        //if(!this->m_first_time) {
            for(auto &u : pObj_Worker) 
                u.reset();            
            pObj_Worker.clear();
            pObj_Worker.shrink_to_fit();
        //}
     
        pObj_Worker.begin();
        
        /* lsblk */
        vector<string> no_hotplug;
        unique_ptr<Hopman::fPipe> fp1( new Hopman::fPipe() );
        vector<string> lsblk_output=fp1->get_output( "lsblk --noheadings --sort PATH -o PATH,HOTPLUG" );
		no_hotplug.begin();	
		for(auto &u : lsblk_output) {
			stringstream ss;
			string path, hotplug;
			ss << u;
			ss >> path >> hotplug;
			// Remove spaces in blanck;
			path.erase(std::remove(path.begin(), path.end(), ' '), path.end());
			hotplug.erase(std::remove(hotplug.begin(), hotplug.end(), ' '), hotplug.end());
			if(stoi(hotplug) == 0)
				no_hotplug.push_back(path);
		}     
                
        /* blkid */
        vector<BlkData> blk_data;
        unique_ptr<Hopman::Svr> svr( new Hopman::Svr( annex.worker_socket ) );
        svr->dispatch( { "0" } );
        unique_ptr<Hopman::fPipe> fp2( new Hopman::fPipe() );
        string cmd = annex.path_to_the_helper + " " + annex.worker_socket;  /* binary + argv[1] */
        vector<string> blkid_output=fp2->get_output( cmd.c_str() );
        if(blkid_output.size()) {
        blk_data.begin();
            for(auto &u : blkid_output) {
                BlkData *item=new BlkData();
                vector<string> tags = { "TYPE", "LABEL", "UUID", "PARTUUID"};
                size_t pos=u.find(":", 1);
                if(pos != -1) {
                    item->path=u.substr(0, pos);
                    for(auto &w : tags) {
                        pos=u.find(w);
                        if(pos != -1) {
                            string str=u.substr(pos);
                            pos=str.find("=");
                            str=str.substr(pos+2); // +2 due to "=\""
                            pos=str.find("\"");
                            str=str.substr(0, pos);
                        
                            if(!w.compare("TYPE"))
                                item->fstype = str;
                            else if(!w.compare("LABEL"))
                                item->label = str;
                            else if(!w.compare("UUID"))
                                item->uuid = str;
                            else if(!w.compare("PARTUUID"))
                                item->partuuid = str;
                        }
                    }
                }
                
                // Is it hotplug?
                bool hotplug = true;
                for(auto &u : no_hotplug)
					if(!(item->path).compare(u))
						hotplug = false;
            
                if(hotplug) 
					blk_data.push_back(*item);
                                
                delete item;                
                tags.clear();
                tags.shrink_to_fit();
            }
        }
        
        no_hotplug.clear();
        no_hotplug.shrink_to_fit();
        
        blkid_output.clear();
        blkid_output.shrink_to_fit();
                    
        if(blk_data.size() == 0) {
            caller->notify();
            m_shall_stop = true;
        }
        
        {
            DIR *dirp;
            struct dirent *E;
        
            vector<string> procmounts;
            procmounts.begin();
            try {
        
                char ch;
                string str;        
                unique_ptr<FILE, decltype(&fclose)> fp(::fopen("/proc/mounts", "r"), &fclose);
                while ((ch = (char)fgetc(fp.get()))!=EOF) {
                    if(ch != '\n') {
                        stringstream s_stream;
                        s_stream << ch;
                        str.append(s_stream.str());
                    }
                    else {
                        procmounts.push_back(str);
                        str.clear();
                    }
                }
            }
    
            catch(exception const& e) {
        
                throw runtime_error("Sorry, there was an error opening the pipe\n");        
            };
  
            // Make the watch directory current
            if( chdir(annex.watch_dir.c_str()) )
            {
                perror(annex.watch_dir.c_str());
                exit(-1);
            }
            
            dirp = opendir(".");
            if(!dirp) {
                fprintf(stderr, _("error reading current directory: %s\n"), strerror(errno));
                exit(EXIT_FAILURE);
            }
                        
            while( (errno=0, E=readdir(dirp)) ) {
                struct stat devstat;

                if( lstat(E->d_name, &devstat) ) {
                    fprintf(stderr, _("[WARN]: create unknown device %s\n"), E->d_name);
                    continue;
                }
                
                string strName = string(E->d_name);
                
                regex_t regex;
                
                // By-pass loopX
                const char *pattern_1 = "^(loop)[0-9]$";  

                int rc = regcomp(&regex, pattern_1, REG_EXTENDED|REG_NOSUB);
                if (rc) {
                    fprintf(stderr, "Could not compile regex: %s\n", pattern_1, strerror(errno) );
                    exit(EXIT_FAILURE);
                }

                rc = regexec(&regex, E->d_name, (size_t)0, NULL, 0);
                if (rc == 0 )
                    continue;
                
                // By-pass srX                   
                const char *pattern_2 = "^(sr)[0-9]$";  

                rc = regcomp(&regex, pattern_2, REG_EXTENDED|REG_NOSUB);
                if (rc) {
                    fprintf(stderr, "Could not compile regex: %s\n", pattern_2, strerror(errno) );
                    exit(EXIT_FAILURE);
                }

                rc = regexec(&regex, E->d_name, (size_t)0, NULL, 0);
                if (rc == 0 )
                    continue;
                    
                // Is it listed by blkid?
                bool listed = false;
                if( blk_data.size() ) {                    
                    for (auto &u : blk_data) {
                        // Is the last character of the item listed by blkid a letter or a number? Disk or partition?
					    string basename = get_basename(u.path);
                        if( !is_partition(strName) ) basename = rm_digit(basename);
                        if( !strName.compare(basename) ) {
                            listed = true;
                            break;
                        }
                    }
                }
                
                if( !listed )
                    continue;
                                
                if( (devstat.st_mode & S_IFMT) == S_IFBLK ) {
                    char namebuf[NAME_MAX+1], *pathname;
                    pathname = (char*)malloc(sizeof(char) * NAME_MAX+1);
                    if (!pathname) {
                        fprintf(stderr, _("Memory assignment failure '%s'\n"), pathname);
                        exit (EXIT_FAILURE);
                    }
                    
                    memset(pathname, 0, NAME_MAX);
                    
				    if(copy_obj.size() > 0) { 
					    vector<shared_ptr<Hopman::BlkNode>>::iterator iter;
                        iter = copy_obj.begin();
                        while(iter < copy_obj.end()) {
							string path = (*iter)->get_pathname();
							string basename = get_basename(path);
							if(!is_partition(strName)) { 
								if( !rm_digit(basename).compare(strName) ) {                      
									if( fs::is_directory(path) ) {
										strcpy(pathname, path.c_str());
										break;
									}
								}
							}
							else {
								for(auto &u : (*iter)->children) {
									string path = u->get_pathname();
									string basename = get_basename(path);
									if( !basename.compare(strName) ) {                          
										if( fs::is_directory(path) ) {
											strcpy(pathname, path.c_str());
											break;
										}
									}
								}
							}
							iter++;
                       }
                   }
                   
                    if(strlen(pathname)==0) {
                        if( !(pathname=devpath(E->d_name, namebuf)) ) {
                            fprintf(stderr, _("cannot find sys path for %s.\n"), E->d_name);
                            continue;
                        }
                    } 
                                        
                    string mountpoint( "" );
                    for(auto &u : procmounts) {
                        vector<string> tokens=BlkTreeWorker::split_line(u, " ");
                        if(!tokens[0].compare( annex.watch_dir + string(E->d_name) )) {
                            mountpoint=tokens[1];
                        }
                    }
                    
                    list = vdev_usb_device_add( list, pathname, mountpoint.c_str(), E->d_ino );
                }
            }
            
            closedir(dirp);
            
            procmounts.clear();
            procmounts.shrink_to_fit();
        }                    
                    
        struct vdev_usb_device_list *aux = list;
        
        int n = 0, m = 0;
        while (aux) {
            char size_str[256] = {0};
            char inode_str[256] = {0};
            
            sprintf(size_str, "%.2f G", aux->device.size);
            sprintf(inode_str, "%d", aux->device.inode);
            
            // Create a new blknode from the item of the list
            auto p = make_shared< Hopman::BlkNode >( aux->device );
            
            // devname
            const char *name=basename( (char*)aux->device.sysfs_path );
            string devname=annex.watch_dir + string(name);
            
            // May or may not be
            bool removed = true;
            
            // devtype = disk? 
            if(!strcmp( aux->device.devtype_str, "disk" )) {
                
                if( blk_data.size() ) {
                    
                    for (auto &u : blk_data) {
                        // Is the last character of the item listed by blkid a letter or a number? Disk or partition?
                        string::iterator iter=(u.path).end();
                        --iter;
                        
                        // Disk...?
                        // Not usual, but it might be listed by blkid even being a disk instead of a partition
                        // If so, get the data 
                        if( isalpha( *iter ) && !(u.path).compare(devname) ) {
                            removed = false;
                            p->set_label(u.label);
                            p->set_fstype(u.fstype);
                            p->set_uuid(u.uuid);
                        }
                        
                        // ... or partition? Are child partitions listed by blkid?
                        // The data is irrelevant here, we only want to stablish the 'removed' variable
                        else if( isdigit( *iter ) ) {
                            string duplicate = u.path;
                            // Remove the number and compare names 
                            duplicate.pop_back();
                            if( !duplicate.compare(devname) )
                                removed = false;
                        }
                    }
                }
                
                // At this point, we know certainty whether or not the device has been removed by software at some time
                if(!removed) {
                    
                    // Is removable?
                    // Read the value in $SYSFS_PATH/removable
                    try {
                        string filename = string(aux->device.sysfs_path) + "/removable";
                        unique_ptr<FILE, decltype(&fclose)> fp(::fopen(filename.c_str(), "r"), &fclose);
                        char ch = (char)fgetc(fp.get());
                        if (ch == '0')
                            p->set_removable(false);
                        else if (ch == '1')
                            p->set_removable(true);
                    }    
    
                    catch(exception const& e) {
            
                        throw runtime_error("Sorry, there was an error opening the pipe\n");        
                    };
                    
                    pObj_Worker.push_back(p);
                    n++;
                }
            } 
            
            // devtype = part?
            else if(!strcmp( aux->device.devtype_str, "partition" )) {
                
                // Is it listed by blkid?
                if( blk_data.size() ) {
                    for(auto &u : blk_data) {
                        if( !(u.path).compare(devname) ) {
							
							int m = 0;
						
							// Remove the number at the end:
							string duplicate = u.path;
							duplicate.pop_back();
						
							while( duplicate.compare( pObj_Worker.at(m)->get_devname() ))
								m++;
								
                            p->parent = pObj_Worker.at(m);
                            
                            // the weak_ptr has to be copied into a shared_ptr before usage
                            if (auto q = p->parent.lock()) {
                                
                                q->children.push_back(p);
                                
                                int dim = q->children.size();
                                q->children.at(dim-1)->set_label(u.label);
                                q->children.at(dim-1)->set_fstype(u.fstype);
                                q->children.at(dim-1)->set_uuid(u.uuid);
                                q->children.at(dim-1)->set_partuuid(u.partuuid);
                                
                            } else {
                     
                                throw runtime_error( "weak pointer is expired\n" );
                            }
                        }
                    }
                }
            }
                    
            aux = aux->next_ptr;
            
        } // while
                
        blk_data.clear();
        blk_data.shrink_to_fit();
        
        if(m_first_time) showList(list);
        
        if (list) 
            list = vdev_usb_device_clear(list);    
            
        free(list);         
                
        if(m_shall_stop) break;

        // Sort by name
        sort(pObj_Worker.begin(), pObj_Worker.end(), compareNames);
            for(auto &u : pObj_Worker) 
                sort( (u->children).begin(), (u->children).end(), compareNames );    
            
        bool shall_notify=false;
        if(pObj_Worker.size() == copy_obj.size()) {
            unsigned i=0;
            while(i<pObj_Worker.size()) {
                if(!compareNodes(pObj_Worker[i], copy_obj[i])) {
                    shall_notify=true;
                    break;
                }
                    
                i++;
            }
        } else shall_notify=true;
            
        if(shall_notify && n)
            caller->notify();
        
        for(auto& u : copy_obj)
            u.reset();            
        copy_obj.clear();
        copy_obj.shrink_to_fit();
        
        copy_obj.begin();
        vector<shared_ptr<Hopman::BlkNode>>::iterator iter;
        iter = pObj_Worker.begin();
        for (; iter < pObj_Worker.end(); iter++)
            copy_obj.push_back(make_shared<Hopman::BlkNode>(**iter));
        
        if(m_first_time)
            this->m_first_time = false;
        
        int length, count = 0;
		char buffer[EVENT_BUF_LEN];

		/*read to determine the event change happens on the monitoring list. Actually this read blocks until the change event occurs*/ 	
		length = read( inot_fd, buffer, EVENT_BUF_LEN );  

		/*checking for error*/
		if ( length < 0 ) {
			perror( "read" );
		}  

		/*actually read return the list of change events happens. Here, read the change event one by one and process it accordingly.*/
		while ( count < length ) {     
			struct inotify_event *event = ( struct inotify_event * ) &buffer[ count ];     
			if ( event->len ) {
				if ( event->mask & IN_CREATE ) {
					if ( event->mask & IN_ISDIR ) {
						printf( "New directory %s created.\n", event->name );
					}
					else {
						printf( "New file %s created.\n", event->name );
					}
				}
				else if ( event->mask & IN_DELETE ) {
					if ( event->mask & IN_ISDIR ) {
						printf( "Directory %s deleted.\n", event->name );
					}
					else {
						printf( "File %s deleted.\n", event->name );
					}
				}
				else if ( event->mask & IN_MODIFY ) {
					printf( "Changed.\n", event->name );
				}
			}
			
			count += EVENT_SIZE + event->len;
		}
	
		/*removing directories from the watch list.*/
		inotify_rm_watch( inot_fd, watch_dirs );
	
		/*closing the INOTIFY instance*/
		close( inot_fd );
         
        this_thread::sleep_for(chrono::milliseconds(1000));
        
    } // while(1)    
        
    for(auto& u : copy_obj)
        u.reset();
                    
    copy_obj.clear();
    copy_obj.shrink_to_fit();

    {
        lock_guard<mutex> lock(m_Mutex);
        m_shall_stop = false;
    }
}

bool BlkTreeWorker::is_partition(string str)
{
    string::iterator iter = str.end();
    --iter;
    if( isdigit(*iter) ) return true;
    else return false;	
}

string BlkTreeWorker::get_basename(string str)
{
	return fs::path(str).filename();	
}

string BlkTreeWorker::rm_digit(string str)
{
	if( is_partition(str) ) str.pop_back();
	return str;
}

void BlkTreeWorker::showList (struct vdev_usb_device_list *l)
{
    struct vdev_usb_device_list *aux = l;
    while(aux) {
        printf("\nVENDOR=%s\n", aux->device.vendor_id);
        printf("MODEL=%s\n", aux->device.model_str);
        printf("TYPE=%s\n", aux->device.type_str);
        printf("DEVTYPE=%s\n", aux->device.devtype_str);
        printf("SIZE=%s\n", aux->device.size_str);
        printf("INODE=%s\n", aux->device.inode_str);
        printf("MTPT=%s\n", aux->device.mtpt_str);
        printf("MAYOR_MINOR=%s\n", aux->device.mayor_minor);
        printf("SYSFS_PATH=%s\n\n", aux->device.sysfs_path);
        aux = aux->next_ptr;
    }
    
    this_thread::sleep_for(chrono::milliseconds(1000));
} 

bool BlkTreeWorker::compareNames(shared_ptr<Hopman::BlkNode> a, shared_ptr<Hopman::BlkNode> b)
{
    return a->get_basename() < b->get_basename();
}

bool BlkTreeWorker::compareNodes(shared_ptr<Hopman::BlkNode> a, shared_ptr<Hopman::BlkNode> b)
{
    if(a->get_inode() == b->get_inode()) {
        vector<string> a_mtpts=a->get_children_mtpt_vector();
        vector<string> b_mtpts=b->get_children_mtpt_vector();
        for(unsigned i=0; i<a_mtpts.size(); i++)
            if(a_mtpts[i].compare(b_mtpts[i])) return false;
    } else return false;
        
    return true;
}

vector<string> BlkTreeWorker::split_line(const string& str, const string& delim)
{
    vector<string> tokens;
    size_t prev = 0, pos = 0;
    do {
        pos = str.find(delim, prev);
        if(pos == string::npos)
            pos = str.length();
        string token = str.substr(prev, pos-prev);
        if(!token.empty())
            tokens.push_back(token);
        prev = pos + delim.length();
    } while (pos < str.length() && prev < str.length());
    
    return tokens;
}

} // namespace Hopman
