/*
 * main.cpp
 * Copyright (C) 2021 Aitor Cuadrado Zubizarreta <aitor_czr@gnuinos.org>
 * 
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * See the LICENSE file. * 
 */


#include <gtkmm.h>
#include <vector>
#include <iostream>
#include <memory>

#include <iostream>
#include <future>
#include <thread>
#include <chrono>

#include <signal.h>
#include <iso646.h>
#include <sys/types.h>
#include <inttypes.h>

#include <X11/Xlib.h>
#include <openssl/sha.h>
#include <stddef.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include "gtkmm/annex.h"
#include "gtkmm/hopman_ui.h"
#include "gtkmm/fpipe.h"
 
using namespace std;

extern "C" {
	
	void sha256_file(char*, char*);
}

/* called when Gtk::Main quits */
bool game_over()
{
    std::cout << "\n\033[1m" << "EXIT!" << "\033[m\n" << std::endl;
    return true;
}

void sig_handler(int signum)
{
	if(signum == SIGUSR1)
		gtk_main_quit();
	else
		cout << "received signal " << signum;
}

void watch_mounts(string mypid)
{
	struct stat st = {0};
	char calc_hash1[65] = {0};

	if (stat("/tmp/hopman1000", &st) == -1) {
		mkdir("/tmp/hopman1000", 0700);
	}
	
	while(1) {
		
		string str = "/tmp/hopman1000/" + mypid;
		char calc_hash2[65] = {0};
		sha256_file( "/proc/mounts", calc_hash2);
		if( strcmp(calc_hash1, calc_hash2) ) {
			struct stat st = {0};
			strcpy(calc_hash1, calc_hash2 );

			// Do some change in /tmp/hopman1000, watched by inotify in the worker thread
			if (stat(str.c_str(), &st) == -1) {
				mkdir(str.c_str(), 0700);
			}
			else {
				rmdir(str.c_str());
			}
		}
		
		this_thread::sleep_for(chrono::milliseconds(1000));
	}
}

void get_xdpy_dim(int *x, int *y)
{
    Display *display;
    Screen *screen;

    /* open a display */	
    try {
		
        display = XOpenDisplay(NULL);
    }
	
    catch(std::exception const& e) {
        
        throw std::runtime_error("Sorry, cannot connect to X server!\n");		
    };

       /* return the number of available screens (unused here) */
    // int count_screens = ScreenCount(display);
       /* Take the first one as reference */
    // screen = ScreenOfDisplay(display, 0);
	
	/* Take de default screen */
    screen = DefaultScreenOfDisplay(display);
    *x = screen->width;
    *y = screen->height;

    /* close the display */
    XCloseDisplay(display);
}


/* *
 *     MAIN                                                 
 */

int main(int argc, char *argv[])
{
    pid_t pid = getpid();	
    string mypid;	
    char *c = nullptr;
    if (asprintf(&c, "%jd", (intmax_t) pid) != -1) {
       mypid = string(c);
       free(c); // cleanup when done.
    } else {	
	   cerr << "Unable to get the pid " << endl;
    }
    
    /* kill previous instances of hopman using the SIGUSR1 signal handler */
    /* 2 seconds trial */
    chrono::system_clock::time_point _2sec_ = chrono::system_clock::now() + chrono::seconds(2);
    promise<int> op_promise;
    future<int> op_future = op_promise.get_future();
	 
    // TODO: before killing the processes, check the data of each one using Jude Nelson's pstat program	
    thread
    (
         [ mypid ]( promise<int> op_promise )
        {	
            int value = 2;  // This should not happen
            			
            unique_ptr<Hopman::fPipe> fp(new Hopman::fPipe());	
            vector<string> output = fp->get_output("ps -A | grep \" hopman$\"");	
            if(output.size() < 1) goto Finish;
            else if(output.size() > 1) {
                const char delim[2] = " ";
                for(unsigned i=0; i<output.size(); i++) {
                    char *token = strtok( (char*)output.at(i).c_str(), delim );
                    if(string(token).compare(mypid)) {
                        errno = 0;
                        char *tmp;
                        intmax_t xmax = strtoimax(token, &tmp, 10);
                        if(errno != 0 or tmp == token or *tmp != '\0' or xmax != (pid_t)xmax)
                            throw runtime_error("Bad PID!\n");
                        else
                            kill( (pid_t)xmax, SIGUSR1 );
                        }
                }
					
                value = 1;
					
                } else
				
                value = 0;  // Run!
Finish:				
            op_promise.set_value_at_thread_exit( value );
        }, 
            move(op_promise)
			
    ).detach();	
		
     if(future_status::ready == op_future.wait_until( _2sec_ ))
    {			
        switch( op_future.get() ) {
            case 0: break;
            case 1: return EXIT_SUCCESS;
            case 2: throw runtime_error("Unable to get the instances of hopman.\n");
            default: return EXIT_FAILURE;
        }
    }
     else
    {
        cout << "Something went wrong\n";
        return EXIT_FAILURE;
    }
    
    /* Register signal handlers */
    struct sigaction sa;
	sa.sa_handler = sig_handler;
	sa.sa_flags = SA_RESTART;
	sigemptyset(&sa.sa_mask);
	sigaction(SIGUSR1, &sa, 0);
    
    std::thread th(watch_mounts, mypid);  // spawn new thread that calls watch_mounts()
    th.detach();    
	
    /* Get the dimensions of the screen */
    int x, y;
    get_xdpy_dim(&x, &y);	
    auto annex = make_unique<Hopman::Annex>(argc, argv, x, y);
	
#if GTK_MAJOR_VERSION == 2
	
    /* Standard boilerplate: initialize the toolkit */
    Gtk::Main kit(argc, argv);
	
    Hopman::Hopman_Ui Hopman_Ui(*annex);
    Gtk::Main::signal_quit().connect (sigc::ptr_fun(&game_over));
    Gtk::Main::run(Hopman_Ui);
    return EXIT_SUCCESS;
	
#elif GTK_MAJOR_VERSION == 3

    auto app = Gtk::Application::create(argc, argv);
    unique_ptr<Hopman::Hopman_Ui> Hopman_Ui = make_unique<Hopman::Hopman_Ui>(*annex);
    return app->run(*Hopman_Ui);
		
#endif
}
