/* See LICENSE file for copyright and license details. */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include "hopman.h"
#include "watch.h"

static partlist_t P = {0, NULL};

/*------------------------------- Creator -----------------------------------*/
void partition_new( ino_t inode, const char *name )
{
  /* reallocate memory for the new number of partitions */
  P.p = realloc( P.p, sizeof(partinfo_t) * (P.np+1) );

  if( P.np && !P.p )
    {
      fprintf( stderr, "%s Cannot store partition data: %s\n",
	       progname, strerror(errno) );
      exit( EXIT_FAILURE );
    }

  /* set the inode and name of the new partition */
  P.p[P.np].inode = inode;
  strncpy(P.p[P.np].name, name, PM_NAME_LEN);
  P.p[P.np].label[0] = P.p[P.np].label[PM_LABEL_LEN] = '\0';
  P.p[P.np].fstype[0] = P.p[P.np].fstype[PM_FSTYPE_LEN] = '\0';
  P.p[P.np].mnt[0] = P.p[P.np].label[PM_MNT_LEN] = '\0';
  P.p[P.np].nmounts = 0;
  
  /* Tell ui we have a new partition */
  ui_create(P.p+P.np);

  /* If we've created the 1st partition, start watching /proc/self/mountinfo */
  if(!P.np) start_periodic(PM_ITIMERVAL);
  
  P.np ++;
  return;
}

/*------------------------------- Destructor --------------------------------*/
void partition_delete( const char *name )
{
  unsigned index;

  /* Find partition */
  for(index=0; index<P.np; index ++) if(!strcmp(P.p[index].name, name)) break;
  if(index == P.np) return;

  /* Tell ui to remove partition from display */
  ui_delete(P.p+index);

  -- P.np;
  
  /* reshuffle the list */
  for( ; index < P.np; index ++) P.p[index] = P.p[index+1];
  
  /* reallocate memory for the new number of partitions */
  P.p = realloc( P.p, sizeof(partinfo_t) * P.np );

  if( P.np && !P.p )
    {
      fprintf( stderr, "%s Error freeing memory, returned by realloc(): %s\n",
	       progname, strerror(errno) );
      exit( EXIT_FAILURE );
    }

  /* If we've deleted the last partition, stop watching /proc/self/mountinfo */
  if(!P.np) stop_periodic();
  
  return;
}

/*------------------------------- Getters -----------------------------------*/
partlist_t partition_get_list(void)
{
  /* for use in mountpoints.c - dangerous, use with care ! */
  return P;
}

unsigned partition_by_name( const char *name )
{
  unsigned index;
  for(index=0; index<P.np; index++)
    {
      if( !strcmp(P.p[index].name, name) ) break;
    }
  return index; /* If partition isn't found, return number of partitions */
}

unsigned partition_by_inode( ino_t inode )
{
  unsigned index;
  for(index=0; index<P.np; index++)
    {
      if( P.p[index].inode == inode ) break;
    }
  return index; /* If partition isn't found, return number of partitions */
}

/*--------------------------------- Setters ---------------------------------*/
void partition_set_label( ino_t inode, const char *label )
{
  unsigned index;
  index = partition_by_inode(inode);
  if(index < P.np)
    {
      strncpy(P.p[index].label, label, PM_LABEL_LEN);
      ui_update(P.p+index);
    }
}
